﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.MultiObjectiveTestFunctions {
  [StorableClass]
  [Item("HypervolumeAnalyzer", "Computes the enclosed Hypervolume between the current front and a given reference Point")]
  public class HypervolumeAnalyzer : MOTFAnalyzer {

    public IValueParameter<DoubleArray> ReferencePointParameter {
      get { return (IValueParameter<DoubleArray>)Parameters["ReferencePoint"]; }
    }

    public IFixedValueParameter<DoubleValue> BestKnownHyperVolumeParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["BestKnownHyperVolume"]; }
    }

    public double BestKnownHyperVolume {
      get { return BestKnownHyperVolumeParameter.Value.Value; }
      set { BestKnownHyperVolumeParameter.Value.Value = value; }
    }

    [StorableConstructor]
    protected HypervolumeAnalyzer(bool deserializing) : base(deserializing) { }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterEventHandlers();
    }

    protected HypervolumeAnalyzer(HypervolumeAnalyzer original, Cloner cloner)
      : base(original, cloner) {
      RegisterEventHandlers();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new HypervolumeAnalyzer(this, cloner);
    }

    public HypervolumeAnalyzer() {
      Parameters.Add(new ValueParameter<DoubleArray>("ReferencePoint", "The reference point for hypervolume calculation"));
      Parameters.Add(new FixedValueParameter<DoubleValue>("BestKnownHyperVolume", "The currently best known hypervolume", new DoubleValue(0)));

      RegisterEventHandlers();
    }

    private void RegisterEventHandlers() {
      ReferencePointParameter.ValueChanged += (o, e) => BestKnownHyperVolume = 0;
    }

    public override IOperation Apply() {
      var results = ResultsParameter.ActualValue;
      var qualities = QualitiesParameter.ActualValue;
      var testFunction = TestFunctionParameter.ActualValue;
      int objectives = qualities[0].Length;

      if (!results.ContainsKey("Hypervolume")) results.Add(new Result("Hypervolume", typeof(DoubleValue)));
      if (!results.ContainsKey("Absolute Distance to BestKnownHypervolume")) results.Add(new Result("Absolute Distance to BestKnownHypervolume", typeof(DoubleValue)));

      double best;
      if (!results.ContainsKey("BestKnownHypervolume")) {
        results.Add(new Result("BestKnownHypervolume", typeof(DoubleValue)));
        best = BestKnownHyperVolumeParameter.Value.Value;
      } else {
        best = ((DoubleValue)(results["BestKnownHypervolume"].Value)).Value;
      }


      IEnumerable<double[]> front = NonDominatedSelect.selectNonDominatedVectors(qualities.Select(q => q.ToArray()), testFunction.Maximization(objectives), true);

      double hv = front.Any() ? Hypervolume.Calculate(front, testFunction.ReferencePoint(objectives), testFunction.Maximization(objectives)) : 0;


      if (double.IsNaN(best) || best < hv) {
        best = hv;
        BestKnownFrontParameter.ActualValue = new DoubleMatrix(MultiObjectiveTestFunctionProblem.To2D(qualities.Select(q => q.ToArray()).ToArray()));
      }

      results["Hypervolume"].Value = new DoubleValue(hv);
      results["BestKnownHypervolume"].Value = new DoubleValue(best);
      results["Absolute Distance to BestKnownHypervolume"].Value = new DoubleValue(best - hv);

      return base.Apply();
    }

  }
}
